# Generated by ORMatic

from __future__ import annotations
from sqlalchemy import Column, ForeignKey, Integer, String, Float, Boolean, DateTime, Enum, JSON
from sqlalchemy.orm import relationship, Mapped, mapped_column, DeclarativeBase
from typing_extensions import Optional, List, Type

import semantic_world.connections
import semantic_world.degree_of_freedom
import semantic_world.geometry
import semantic_world.orm.model
import semantic_world.prefixed_name
import semantic_world.spatial_types.spatial_types
import semantic_world.views.views
import semantic_world.world
import semantic_world.world_entity


from ormatic.dao import DataAccessObject
from ormatic.custom_types import TypeType

class Base(DeclarativeBase):
    type_mappings = {
        Type: TypeType,
    }


class QuaternionMappingDAO(Base, DataAccessObject[semantic_world.orm.model.QuaternionMapping]):
    __tablename__ = 'QuaternionMappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    x: Mapped[float]
    y: Mapped[float]
    z: Mapped[float]
    w: Mapped[float]


    reference_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    reference_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[reference_frame_id], post_update=True)


class ShapeDAO(Base, DataAccessObject[semantic_world.geometry.Shape]):
    __tablename__ = 'ShapeDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)


    polymorphic_type: Mapped[str] = mapped_column(String(255), nullable=False)

    origin_id: Mapped[int] = mapped_column(ForeignKey('TransformationMatrixMappingDAO.id', use_alter=True), nullable=True)
    bodydao_visual_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)
    bodydao_collision_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    origin: Mapped[TransformationMatrixMappingDAO] = relationship('TransformationMatrixMappingDAO', uselist=False, foreign_keys=[origin_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_on': 'polymorphic_type',
        'polymorphic_identity': 'ShapeDAO',
    }

class Point3MappingDAO(Base, DataAccessObject[semantic_world.orm.model.Point3Mapping]):
    __tablename__ = 'Point3MappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    x: Mapped[float]
    y: Mapped[float]
    z: Mapped[float]


    reference_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    reference_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[reference_frame_id], post_update=True)


class ScaleDAO(Base, DataAccessObject[semantic_world.geometry.Scale]):
    __tablename__ = 'ScaleDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    x: Mapped[float]
    y: Mapped[float]
    z: Mapped[float]





class BoundingBoxCollectionDAO(Base, DataAccessObject[semantic_world.geometry.BoundingBoxCollection]):
    __tablename__ = 'BoundingBoxCollectionDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)




    bounding_boxes: Mapped[List[BoundingBoxDAO]] = relationship('BoundingBoxDAO', foreign_keys='[BoundingBoxDAO.boundingboxcollectiondao_bounding_boxes_id]', post_update=True)


class WorldEntityDAO(Base, DataAccessObject[semantic_world.world_entity.WorldEntity]):
    __tablename__ = 'WorldEntityDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)


    polymorphic_type: Mapped[str] = mapped_column(String(255), nullable=False)

    name_id: Mapped[int] = mapped_column(ForeignKey('PrefixedNameDAO.id', use_alter=True), nullable=True)

    name: Mapped[PrefixedNameDAO] = relationship('PrefixedNameDAO', uselist=False, foreign_keys=[name_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_on': 'polymorphic_type',
        'polymorphic_identity': 'WorldEntityDAO',
    }

class WorldMappingDAO(Base, DataAccessObject[semantic_world.orm.model.WorldMapping]):
    __tablename__ = 'WorldMappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)




    bodies: Mapped[List[BodyDAO]] = relationship('BodyDAO', foreign_keys='[BodyDAO.worldmappingdao_bodies_id]', post_update=True)
    connections: Mapped[List[ConnectionDAO]] = relationship('ConnectionDAO', foreign_keys='[ConnectionDAO.worldmappingdao_connections_id]', post_update=True)
    views: Mapped[List[ViewDAO]] = relationship('ViewDAO', foreign_keys='[ViewDAO.worldmappingdao_views_id]', post_update=True)
    degrees_of_freedom: Mapped[List[DegreeOfFreedomMappingDAO]] = relationship('DegreeOfFreedomMappingDAO', foreign_keys='[DegreeOfFreedomMappingDAO.worldmappingdao_degrees_of_freedom_id]', post_update=True)


class ColorDAO(Base, DataAccessObject[semantic_world.geometry.Color]):
    __tablename__ = 'ColorDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    R: Mapped[float]
    G: Mapped[float]
    B: Mapped[float]
    A: Mapped[float]





class Vector3MappingDAO(Base, DataAccessObject[semantic_world.orm.model.Vector3Mapping]):
    __tablename__ = 'Vector3MappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    x: Mapped[float]
    y: Mapped[float]
    z: Mapped[float]


    reference_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    reference_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[reference_frame_id], post_update=True)


class BoundingBoxDAO(Base, DataAccessObject[semantic_world.geometry.BoundingBox]):
    __tablename__ = 'BoundingBoxDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    min_x: Mapped[float]
    min_y: Mapped[float]
    min_z: Mapped[float]
    max_x: Mapped[float]
    max_y: Mapped[float]
    max_z: Mapped[float]


    boundingboxcollectiondao_bounding_boxes_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BoundingBoxCollectionDAO.id', use_alter=True), nullable=True)



class TransformationMatrixMappingDAO(Base, DataAccessObject[semantic_world.orm.model.TransformationMatrixMapping]):
    __tablename__ = 'TransformationMatrixMappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)



    position_id: Mapped[int] = mapped_column(ForeignKey('Point3MappingDAO.id', use_alter=True), nullable=True)
    rotation_id: Mapped[int] = mapped_column(ForeignKey('QuaternionMappingDAO.id', use_alter=True), nullable=True)
    reference_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)
    child_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    position: Mapped[Point3MappingDAO] = relationship('Point3MappingDAO', uselist=False, foreign_keys=[position_id], post_update=True)
    rotation: Mapped[QuaternionMappingDAO] = relationship('QuaternionMappingDAO', uselist=False, foreign_keys=[rotation_id], post_update=True)
    reference_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[reference_frame_id], post_update=True)
    child_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[child_frame_id], post_update=True)


class PrefixedNameDAO(Base, DataAccessObject[semantic_world.prefixed_name.PrefixedName]):
    __tablename__ = 'PrefixedNameDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)


    name: Mapped[str] = mapped_column(String(255), nullable=False)
    prefix: Mapped[Optional[str]] = mapped_column(String(255), nullable=True)




class RotationMatrixMappingDAO(Base, DataAccessObject[semantic_world.orm.model.RotationMatrixMapping]):
    __tablename__ = 'RotationMatrixMappingDAO'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)



    rotation_id: Mapped[int] = mapped_column(ForeignKey('QuaternionMappingDAO.id', use_alter=True), nullable=True)
    reference_frame_id: Mapped[Optional[int]] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    rotation: Mapped[QuaternionMappingDAO] = relationship('QuaternionMappingDAO', uselist=False, foreign_keys=[rotation_id], post_update=True)
    reference_frame: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[reference_frame_id], post_update=True)


class PrimitiveDAO(ShapeDAO, DataAccessObject[semantic_world.geometry.Primitive]):
    __tablename__ = 'PrimitiveDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ShapeDAO.id), primary_key=True)



    color_id: Mapped[int] = mapped_column(ForeignKey('ColorDAO.id', use_alter=True), nullable=True)

    color: Mapped[ColorDAO] = relationship('ColorDAO', uselist=False, foreign_keys=[color_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'PrimitiveDAO',
        'inherit_condition': id == ShapeDAO.id,
    }

class MeshDAO(ShapeDAO, DataAccessObject[semantic_world.geometry.Mesh]):
    __tablename__ = 'MeshDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ShapeDAO.id), primary_key=True)


    filename: Mapped[str] = mapped_column(String(255), nullable=False)

    scale_id: Mapped[int] = mapped_column(ForeignKey('ScaleDAO.id', use_alter=True), nullable=True)

    scale: Mapped[ScaleDAO] = relationship('ScaleDAO', uselist=False, foreign_keys=[scale_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'MeshDAO',
        'inherit_condition': id == ShapeDAO.id,
    }

class BodyDAO(WorldEntityDAO, DataAccessObject[semantic_world.world_entity.Body]):
    __tablename__ = 'BodyDAO'

    id: Mapped[int] = mapped_column(ForeignKey(WorldEntityDAO.id), primary_key=True)

    index: Mapped[Optional[int]]


    worldmappingdao_bodies_id: Mapped[Optional[int]] = mapped_column(ForeignKey('WorldMappingDAO.id', use_alter=True), nullable=True)

    visual: Mapped[List[ShapeDAO]] = relationship('ShapeDAO', foreign_keys='[ShapeDAO.bodydao_visual_id]', post_update=True)
    collision: Mapped[List[ShapeDAO]] = relationship('ShapeDAO', foreign_keys='[ShapeDAO.bodydao_collision_id]', post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'BodyDAO',
        'inherit_condition': id == WorldEntityDAO.id,
    }

class DegreeOfFreedomMappingDAO(WorldEntityDAO, DataAccessObject[semantic_world.orm.model.DegreeOfFreedomMapping]):
    __tablename__ = 'DegreeOfFreedomMappingDAO'

    id: Mapped[int] = mapped_column(ForeignKey(WorldEntityDAO.id), primary_key=True)


    lower_limits: Mapped[List[float]] = mapped_column(JSON, nullable=False)
    upper_limits: Mapped[List[float]] = mapped_column(JSON, nullable=False)

    worldmappingdao_degrees_of_freedom_id: Mapped[Optional[int]] = mapped_column(ForeignKey('WorldMappingDAO.id', use_alter=True), nullable=True)
    name_id: Mapped[int] = mapped_column(ForeignKey('PrefixedNameDAO.id', use_alter=True), nullable=True)

    name: Mapped[PrefixedNameDAO] = relationship('PrefixedNameDAO', uselist=False, foreign_keys=[name_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'DegreeOfFreedomMappingDAO',
        'inherit_condition': id == WorldEntityDAO.id,
    }

class ConnectionDAO(WorldEntityDAO, DataAccessObject[semantic_world.world_entity.Connection]):
    __tablename__ = 'ConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(WorldEntityDAO.id), primary_key=True)



    worldmappingdao_connections_id: Mapped[Optional[int]] = mapped_column(ForeignKey('WorldMappingDAO.id', use_alter=True), nullable=True)
    parent_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)
    child_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)
    origin_expression_id: Mapped[int] = mapped_column(ForeignKey('TransformationMatrixMappingDAO.id', use_alter=True), nullable=True)

    parent: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[parent_id], post_update=True)
    child: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[child_id], post_update=True)
    origin_expression: Mapped[TransformationMatrixMappingDAO] = relationship('TransformationMatrixMappingDAO', uselist=False, foreign_keys=[origin_expression_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'ConnectionDAO',
        'inherit_condition': id == WorldEntityDAO.id,
    }

class ViewDAO(WorldEntityDAO, DataAccessObject[semantic_world.world_entity.View]):
    __tablename__ = 'ViewDAO'

    id: Mapped[int] = mapped_column(ForeignKey(WorldEntityDAO.id), primary_key=True)



    worldmappingdao_views_id: Mapped[Optional[int]] = mapped_column(ForeignKey('WorldMappingDAO.id', use_alter=True), nullable=True)


    __mapper_args__ = {
        'polymorphic_identity': 'ViewDAO',
        'inherit_condition': id == WorldEntityDAO.id,
    }

class BoxDAO(PrimitiveDAO, DataAccessObject[semantic_world.geometry.Box]):
    __tablename__ = 'BoxDAO'

    id: Mapped[int] = mapped_column(ForeignKey(PrimitiveDAO.id), primary_key=True)



    scale_id: Mapped[int] = mapped_column(ForeignKey('ScaleDAO.id', use_alter=True), nullable=True)

    scale: Mapped[ScaleDAO] = relationship('ScaleDAO', uselist=False, foreign_keys=[scale_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'BoxDAO',
        'inherit_condition': id == PrimitiveDAO.id,
    }

class SphereDAO(PrimitiveDAO, DataAccessObject[semantic_world.geometry.Sphere]):
    __tablename__ = 'SphereDAO'

    id: Mapped[int] = mapped_column(ForeignKey(PrimitiveDAO.id), primary_key=True)

    radius: Mapped[float]




    __mapper_args__ = {
        'polymorphic_identity': 'SphereDAO',
        'inherit_condition': id == PrimitiveDAO.id,
    }

class CylinderDAO(PrimitiveDAO, DataAccessObject[semantic_world.geometry.Cylinder]):
    __tablename__ = 'CylinderDAO'

    id: Mapped[int] = mapped_column(ForeignKey(PrimitiveDAO.id), primary_key=True)

    width: Mapped[float]
    height: Mapped[float]




    __mapper_args__ = {
        'polymorphic_identity': 'CylinderDAO',
        'inherit_condition': id == PrimitiveDAO.id,
    }

class ActiveConnectionDAO(ConnectionDAO, DataAccessObject[semantic_world.connections.ActiveConnection]):
    __tablename__ = 'ActiveConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ConnectionDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'ActiveConnectionDAO',
        'inherit_condition': id == ConnectionDAO.id,
    }

class FixedConnectionDAO(ConnectionDAO, DataAccessObject[semantic_world.connections.FixedConnection]):
    __tablename__ = 'FixedConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ConnectionDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'FixedConnectionDAO',
        'inherit_condition': id == ConnectionDAO.id,
    }

class PassiveConnectionDAO(ConnectionDAO, DataAccessObject[semantic_world.connections.PassiveConnection]):
    __tablename__ = 'PassiveConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ConnectionDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'PassiveConnectionDAO',
        'inherit_condition': id == ConnectionDAO.id,
    }

class RootedViewDAO(ViewDAO, DataAccessObject[semantic_world.world_entity.RootedView]):
    __tablename__ = 'RootedViewDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)



    root_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    root: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[root_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'RootedViewDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class FridgeDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Fridge]):
    __tablename__ = 'FridgeDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)



    body_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    body: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[body_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'FridgeDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class TableDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Table]):
    __tablename__ = 'TableDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)



    top_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    top: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[top_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'TableDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class HandleDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Handle]):
    __tablename__ = 'HandleDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)



    body_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    body: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[body_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'HandleDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class ComponentsDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Components]):
    __tablename__ = 'ComponentsDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'ComponentsDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class FurnitureDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Furniture]):
    __tablename__ = 'FurnitureDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'FurnitureDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class ContainerDAO(ViewDAO, DataAccessObject[semantic_world.views.views.Container]):
    __tablename__ = 'ContainerDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ViewDAO.id), primary_key=True)



    body_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)

    body: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[body_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'ContainerDAO',
        'inherit_condition': id == ViewDAO.id,
    }

class PrismaticConnectionDAO(ActiveConnectionDAO, DataAccessObject[semantic_world.connections.PrismaticConnection]):
    __tablename__ = 'PrismaticConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ActiveConnectionDAO.id), primary_key=True)

    multiplier: Mapped[float]
    offset: Mapped[float]


    axis_id: Mapped[int] = mapped_column(ForeignKey('Vector3MappingDAO.id', use_alter=True), nullable=True)
    dof_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)

    axis: Mapped[Vector3MappingDAO] = relationship('Vector3MappingDAO', uselist=False, foreign_keys=[axis_id], post_update=True)
    dof: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[dof_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'PrismaticConnectionDAO',
        'inherit_condition': id == ActiveConnectionDAO.id,
    }

class OmniDriveDAO(ActiveConnectionDAO, DataAccessObject[semantic_world.connections.OmniDrive]):
    __tablename__ = 'OmniDriveDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ActiveConnectionDAO.id), primary_key=True)

    translation_velocity_limits: Mapped[float]
    rotation_velocity_limits: Mapped[float]


    x_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    y_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    z_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    roll_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    pitch_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    yaw_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    x_vel_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    y_vel_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)

    x: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[x_id], post_update=True)
    y: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[y_id], post_update=True)
    z: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[z_id], post_update=True)
    roll: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[roll_id], post_update=True)
    pitch: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[pitch_id], post_update=True)
    yaw: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[yaw_id], post_update=True)
    x_vel: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[x_vel_id], post_update=True)
    y_vel: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[y_vel_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'OmniDriveDAO',
        'inherit_condition': id == ActiveConnectionDAO.id,
    }

class RevoluteConnectionDAO(ActiveConnectionDAO, DataAccessObject[semantic_world.connections.RevoluteConnection]):
    __tablename__ = 'RevoluteConnectionDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ActiveConnectionDAO.id), primary_key=True)

    multiplier: Mapped[float]
    offset: Mapped[float]


    axis_id: Mapped[int] = mapped_column(ForeignKey('Vector3MappingDAO.id', use_alter=True), nullable=True)
    dof_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)

    axis: Mapped[Vector3MappingDAO] = relationship('Vector3MappingDAO', uselist=False, foreign_keys=[axis_id], post_update=True)
    dof: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[dof_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'RevoluteConnectionDAO',
        'inherit_condition': id == ActiveConnectionDAO.id,
    }

class Connection6DoFDAO(PassiveConnectionDAO, DataAccessObject[semantic_world.connections.Connection6DoF]):
    __tablename__ = 'Connection6DoFDAO'

    id: Mapped[int] = mapped_column(ForeignKey(PassiveConnectionDAO.id), primary_key=True)



    x_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    y_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    z_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    qx_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    qy_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    qz_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)
    qw_id: Mapped[int] = mapped_column(ForeignKey('DegreeOfFreedomMappingDAO.id', use_alter=True), nullable=True)

    x: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[x_id], post_update=True)
    y: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[y_id], post_update=True)
    z: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[z_id], post_update=True)
    qx: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[qx_id], post_update=True)
    qy: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[qy_id], post_update=True)
    qz: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[qz_id], post_update=True)
    qw: Mapped[DegreeOfFreedomMappingDAO] = relationship('DegreeOfFreedomMappingDAO', uselist=False, foreign_keys=[qw_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'Connection6DoFDAO',
        'inherit_condition': id == PassiveConnectionDAO.id,
    }

class EnvironmentViewDAO(RootedViewDAO, DataAccessObject[semantic_world.world_entity.EnvironmentView]):
    __tablename__ = 'EnvironmentViewDAO'

    id: Mapped[int] = mapped_column(ForeignKey(RootedViewDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'EnvironmentViewDAO',
        'inherit_condition': id == RootedViewDAO.id,
    }

class DrawerDAO(ComponentsDAO, DataAccessObject[semantic_world.views.views.Drawer]):
    __tablename__ = 'DrawerDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ComponentsDAO.id), primary_key=True)



    container_id: Mapped[int] = mapped_column(ForeignKey('ContainerDAO.id', use_alter=True), nullable=True)
    handle_id: Mapped[int] = mapped_column(ForeignKey('HandleDAO.id', use_alter=True), nullable=True)
    cabinetdao_drawers_id: Mapped[Optional[int]] = mapped_column(ForeignKey('CabinetDAO.id', use_alter=True), nullable=True)

    container: Mapped[ContainerDAO] = relationship('ContainerDAO', uselist=False, foreign_keys=[container_id], post_update=True)
    handle: Mapped[HandleDAO] = relationship('HandleDAO', uselist=False, foreign_keys=[handle_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'DrawerDAO',
        'inherit_condition': id == ComponentsDAO.id,
    }

class DoorDAO(ComponentsDAO, DataAccessObject[semantic_world.views.views.Door]):
    __tablename__ = 'DoorDAO'

    id: Mapped[int] = mapped_column(ForeignKey(ComponentsDAO.id), primary_key=True)



    body_id: Mapped[int] = mapped_column(ForeignKey('BodyDAO.id', use_alter=True), nullable=True)
    handle_id: Mapped[int] = mapped_column(ForeignKey('HandleDAO.id', use_alter=True), nullable=True)
    wardrobedao_doors_id: Mapped[Optional[int]] = mapped_column(ForeignKey('WardrobeDAO.id', use_alter=True), nullable=True)

    body: Mapped[BodyDAO] = relationship('BodyDAO', uselist=False, foreign_keys=[body_id], post_update=True)
    handle: Mapped[HandleDAO] = relationship('HandleDAO', uselist=False, foreign_keys=[handle_id], post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'DoorDAO',
        'inherit_condition': id == ComponentsDAO.id,
    }

class CupboardDAO(FurnitureDAO, DataAccessObject[semantic_world.views.views.Cupboard]):
    __tablename__ = 'CupboardDAO'

    id: Mapped[int] = mapped_column(ForeignKey(FurnitureDAO.id), primary_key=True)





    __mapper_args__ = {
        'polymorphic_identity': 'CupboardDAO',
        'inherit_condition': id == FurnitureDAO.id,
    }

class WardrobeDAO(CupboardDAO, DataAccessObject[semantic_world.views.views.Wardrobe]):
    __tablename__ = 'WardrobeDAO'

    id: Mapped[int] = mapped_column(ForeignKey(CupboardDAO.id), primary_key=True)




    doors: Mapped[List[DoorDAO]] = relationship('DoorDAO', foreign_keys='[DoorDAO.wardrobedao_doors_id]', post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'WardrobeDAO',
        'inherit_condition': id == CupboardDAO.id,
    }

class CabinetDAO(CupboardDAO, DataAccessObject[semantic_world.views.views.Cabinet]):
    __tablename__ = 'CabinetDAO'

    id: Mapped[int] = mapped_column(ForeignKey(CupboardDAO.id), primary_key=True)



    container_id: Mapped[int] = mapped_column(ForeignKey('ContainerDAO.id', use_alter=True), nullable=True)

    container: Mapped[ContainerDAO] = relationship('ContainerDAO', uselist=False, foreign_keys=[container_id], post_update=True)
    drawers: Mapped[List[DrawerDAO]] = relationship('DrawerDAO', foreign_keys='[DrawerDAO.cabinetdao_drawers_id]', post_update=True)

    __mapper_args__ = {
        'polymorphic_identity': 'CabinetDAO',
        'inherit_condition': id == CupboardDAO.id,
    }




import atexit
import threading
import time

import numpy as np
from builtin_interfaces.msg import Duration
from geometry_msgs.msg import Vector3, Point, PoseStamped, Quaternion, Pose
from scipy.spatial.transform import Rotation
from std_msgs.msg import ColorRGBA
from visualization_msgs.msg import Marker, MarkerArray

from ..geometry import Mesh, Box, Sphere, Cylinder, Primitive
from ..world import World

class VizMarkerPublisher:
    """
    Publishes an Array of visualization marker which represent the situation in the World
    """

    def __init__(self, world: World, node, topic_name="/semworld/viz_marker", interval=0.1, reference_frame="map"):
        """
        The Publisher creates an Array of Visualization marker with a Marker for each body in the World and publishes
        it to the given topic name at a fixed interval. The publisher automatically stops publishing when the process
        is killed.

        :param world: The World to which the Visualization Marker should be published.
        :param node: The ROS2 node that will be used to publish the visualization marker.
        :param topic_name: The name of the topic to which the Visualization Marker should be published.
        :param interval: The interval at which the visualization marker should be published, in seconds.
        :param reference_frame: The reference frame of the visualization marker.
        """
        self.interval = interval
        self.reference_frame = reference_frame
        self.world = world
        self.node = node

        self.pub = self.node.create_publisher(MarkerArray, topic_name, 10)

        self.thread = threading.Thread(target=self._publish, name="VizMarkerPublisher")
        self.kill_event = threading.Event()

        self.thread.start()
        atexit.register(self._stop_publishing)

    def _publish(self) -> None:
        """
        Constantly publishes the Marker Array. To the given topic name at a fixed rate.
        """
        while not self.kill_event.is_set():
            marker_array = self._make_marker_array()
            self.pub.publish(marker_array)
            time.sleep(self.interval)

    def _make_marker_array(self) -> MarkerArray:
        """
        Creates the Marker Array to be published. There is one Marker for link for each object in the Array, each Object
        creates a name space in the visualization Marker. The type of Visualization Marker is decided by the collision
        tag of the URDF.

        :return: An Array of Visualization Marker
        """
        marker_array = MarkerArray()
        for body in self.world.bodies:
            for i, collision in enumerate(body.collision):
                msg = Marker()
                msg.header.frame_id = self.reference_frame
                msg.ns = body.name.name
                msg.id = i
                msg.action = Marker.ADD
                msg.pose = self.transform_to_pose(self.world.compute_forward_kinematics_np(self.world.root, body) @ collision.origin.to_np())
                msg.color = body.color if isinstance(body, Primitive) else ColorRGBA(r=1.0, g=1.0, b=1.0, a=1.0)
                msg.lifetime = Duration(sec=1)

                if isinstance(collision, Mesh):
                    msg.type = Marker.MESH_RESOURCE
                    msg.mesh_resource = "file://" + collision.filename
                    msg.scale = Vector3(x=float(collision.scale.x), y=float(collision.scale.y), z=float(collision.scale.z))
                    msg.mesh_use_embedded_materials = True
                elif isinstance(collision, Cylinder):
                    msg.type = Marker.CYLINDER
                    msg.scale = Vector3(x=float(collision.width), y=float(collision.width), z=float(collision.height))
                elif isinstance(collision, Box):
                    msg.type = Marker.CUBE
                    msg.scale = Vector3(x=float(collision.scale.x), y=float(collision.scale.y), z=float(collision.scale.z))
                elif isinstance(collision, Sphere):
                    msg.type = Marker.SPHERE
                    msg.scale = Vector3(x=float(collision.radius * 2), y=float(collision.radius * 2), z=float(collision.radius * 2))

                marker_array.markers.append(msg)
        return marker_array

    def _stop_publishing(self) -> None:
        """
        Stops the publishing of the Visualization Marker update by setting the kill event and collecting the thread.
        """
        self.kill_event.set()
        self.thread.join()

    @staticmethod
    def transform_to_pose(transform: np.ndarray) -> Pose:
        """
        Converts a 4x4 transformation matrix to a PoseStamped message.

        :param transform: The transformation matrix to convert.
        :return: A PoseStamped message.
        """
        pose = Pose()
        pose.position = Point(**dict(zip(["x", "y", "z"], transform[:3, 3])))
        pose.orientation = Quaternion(**dict(zip(["x", "y", "z", "w"], Rotation.from_matrix(transform[:3, :3]).as_quat())))
        return pose




import logging
import os
import sys
import unittest

import pytest
from numpy.ma.testutils import assert_equal

from semantic_world.reasoner import WorldReasoner

try:
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
    from PyQt6.QtWidgets import QApplication
except ImportError as e:
    logging.debug(e)
    QApplication = None
    RDRCaseViewer = None

from typing_extensions import Type, Optional, Callable

from semantic_world.adapters.urdf import URDFParser
from semantic_world.views import *

try:
    from semantic_world.world_rdr import world_rdr
except ImportError as e:
    world_rdr = None
from semantic_world.world import World

@dataclass
class TestView(View):
    """
    A Generic View for multiple bodies.
    """
    _private_body: Body = field(default=Body)
    body_list: List[Body] = field(default_factory=list, hash=False)
    views: List[View] = field(default_factory=list, hash=False)
    root_body_1: Body = field(default=Body)
    root_body_2: Body = field(default=Body)
    tip_body_1: Body = field(default=Body)
    tip_body_2: Body = field(default=Body)

    def add_body(self, body: Body):
        self.body_list.append(body)
        body._views.add(self)

    def add_view(self, view: View):
        self.views.append(view)
        view._views.add(self)

    @property
    def chain(self) -> list[Body]:
        """
        Returns itself as a kinematic chain.
        """
        return self._world.compute_chain_of_bodies(self.root_body_1, self.tip_body_1)

    @property
    def _private_chain(self) -> list[Body]:
        """
        Returns itself as a kinematic chain.
        """
        return self._world.compute_chain_of_bodies(self.root_body_2, self.tip_body_2)

    def __hash__(self):
        """
        Custom hash function to ensure that the view is hashable.
        """
        return hash((self._private_body, tuple(self.body_list), tuple(self.views),
                     self.root_body_1, self.root_body_2, self.tip_body_1, self.tip_body_2))


class ViewTestCase(unittest.TestCase):
    """
    **Important**:
    ===============
    If use_gui is set to False, use the command line interface to test the views.

    e.g. from the terminal while at the root of the repository, run:

    cd test/test_views && python -m pytest test_views.py

    ===============
    OR if you want to run only the kitchen views test:

    cd test/test_views && python -m pytest -k "test_kitchen_views"

    """
    urdf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf")
    kitchen = os.path.join(urdf_dir, "kitchen-small.urdf")
    apartment = os.path.join(urdf_dir, "apartment.urdf")
    main_dir = os.path.join(os.path.dirname(__file__), '..', '..')
    views_dir = os.path.join(main_dir, "src/semantic_world/views")
    views_rdr_model_name = "world_rdr"
    test_dir = os.path.join(main_dir, 'tests')
    expert_answers_dir = os.path.join(test_dir, "test_expert_answers")
    app: Optional[QApplication] = None
    viewer: Optional[RDRCaseViewer] = None
    use_gui: bool = False

    @classmethod
    def setUpClass(cls):
        cls.kitchen_world = cls.get_kitchen_world()
        cls.apartment_world = cls.get_apartment_world()
        if RDRCaseViewer is not None and QApplication is not None and cls.use_gui:
            cls.app = QApplication(sys.argv)
            cls.viewer = RDRCaseViewer()

    def test_aggregate_bodies(self):
        world_view = TestView(_world=self.kitchen_world)

        # Test bodies added to a private dataclass field are not aggregated
        world_view._private_body = self.kitchen_world.bodies[0]

        # Test aggregation of bodies added in custom properties
        world_view.root_body_1 = self.kitchen_world.bodies[1]
        world_view.tip_body_1 = self.kitchen_world.bodies[4]

        # Test aggregation of normal dataclass field
        body_subset = self.kitchen_world.bodies[5:10]
        [world_view.add_body(body) for body in body_subset]

        # Test aggregation of bodies in a new as well as a nested view
        view1 = TestView()
        view1_subset = self.kitchen_world.bodies[10:18]
        [view1.add_body(body) for body in view1_subset]

        view2 = TestView()
        view2_subset = self.kitchen_world.bodies[20:]
        [view2.add_body(body) for body in view2_subset]

        view1.add_view(view2)
        world_view.add_view(view1)

        # Test that bodies added in a custom private property are not aggregated
        world_view.root_body_2 = self.kitchen_world.bodies[18]
        world_view.tip_body_2 = self.kitchen_world.bodies[20]

        # The aggregation should not include the private dataclass field body or the body added exclusively in the private property
        assert_equal(world_view.bodies, set(self.kitchen_world.bodies) - {self.kitchen_world.bodies[0], self.kitchen_world.bodies[19]})

    def test_handle_view(self):
        self.fit_rules_for_a_view_in_apartment(Handle, scenario=self.test_handle_view)

    def test_container_view(self):
        self.fit_rules_for_a_view_in_apartment(Container, scenario=self.test_container_view)

    def test_drawer_view(self):
        self.fit_rules_for_a_view_in_apartment(Drawer, scenario=self.test_drawer_view)

    def test_cabinet_view(self):
        self.fit_rules_for_a_view_in_apartment(Cabinet, scenario=self.test_cabinet_view)

    def test_door_view(self):
        self.fit_rules_for_a_view_in_apartment(Door, scenario=self.test_door_view)

    def test_fridge_view(self):
        self.fit_rules_for_a_view_in_kitchen(Fridge, scenario=self.test_fridge_view, update_existing_views=False)

    @unittest.skip("Skipping test for wardrobe view as it requires user input")
    def test_wardrobe_view(self):
        self.fit_rules_for_a_view_in_apartment(Wardrobe, scenario=self.test_wardrobe_view)

    @pytest.mark.skipif(world_rdr is None, reason="requires world_rdr")
    def test_generated_views(self):
        found_views = world_rdr.classify(self.kitchen_world)["views"]

        drawer_container_names = [v.body.name.name for v in found_views if isinstance(v, Container)]
        self.assertTrue(len(drawer_container_names) == 14)

    @pytest.mark.order("second_to_last")
    def test_apartment_views(self):
        world_reasoner = WorldReasoner(self.apartment_world)
        world_reasoner.fit_views([Handle, Container, Drawer, Cabinet],
                                 world_factory=self.get_apartment_world, scenario=self.test_apartment_views)

        found_views = world_reasoner.infer_views()

        drawer_container_names = [v.body.name.name for v in found_views if isinstance(v, Container)]

        self.assertTrue(len(drawer_container_names) == 19)

    @pytest.mark.order("last")
    def test_kitchen_views(self):
        world_reasoner = WorldReasoner(self.kitchen_world)
        world_reasoner.fit_views([Handle, Container, Drawer, Cabinet],
                                 world_factory=self.get_kitchen_world, scenario=self.test_kitchen_views)

        found_views = world_reasoner.infer_views()

        drawer_container_names = [v.body.name.name for v in found_views if isinstance(v, Container)]

        self.assertTrue(len(drawer_container_names) == 14)

    @classmethod
    def get_kitchen_world(cls) -> World:
        """
        Return the kitchen world parsed from the URDF file.
        """
        parser = URDFParser(cls.kitchen)
        world = parser.parse()
        world.validate()
        return world

    @classmethod
    def get_apartment_world(cls) -> World:
        """
        Return the apartment world parsed from the URDF file.
        """
        parser = URDFParser(cls.apartment)
        world = parser.parse()
        world.validate()
        return world

    def fit_rules_for_a_view_in_kitchen(self, view_type: Type[View], update_existing_views: bool = False,
                                        scenario: Optional[Callable] = None) -> None:
        """
        Template method to test a specific view type in the kitchen world.

        :param view_type: The type of view to fit and assert.
        :param update_existing_views: If True, existing views will be updated with new rules, else they will be skipped.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        self.fit_rules_for_a_view_and_assert(self.kitchen_world, view_type, update_existing_views=update_existing_views,
                                             world_factory=self.get_kitchen_world, scenario=scenario)

    def fit_rules_for_a_view_in_apartment(self, view_type: Type[View], update_existing_views: bool = False,
                                          scenario: Optional[Callable] = None) -> None:
        """
        Template method to test a specific view type in the apartment world.

        :param view_type: The type of view to fit and assert.
        :param update_existing_views: If True, existing views will be updated with new rules, else they will be skipped.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        self.fit_rules_for_a_view_and_assert(self.apartment_world, view_type,
                                             update_existing_views=update_existing_views,
                                             world_factory=self.get_apartment_world, scenario=scenario)

    @staticmethod
    def fit_rules_for_a_view_and_assert(world: World, view_type: Type[View], update_existing_views: bool = False,
                                        world_factory: Optional[Callable] = None,
                                        scenario: Optional[Callable] = None) -> None:
        """
        Template method to test a specific view type in the given world.

        :param world: The world to fit the view to.
        :param view_type: The type of view to fit and assert.
        :param update_existing_views: If True, existing views will be updated with new rules, else they will be skipped.
        :param world_factory: Optional callable that can be used to recreate the world object.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        world_reasoner = WorldReasoner(world)
        world_reasoner.fit_views([view_type], update_existing_views=update_existing_views,
                                 world_factory=world_factory, scenario=scenario)

        found_views = world_reasoner.infer_views()

        assert any(isinstance(v, view_type) for v in found_views)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, TYPE_CHECKING, Union

import numpy as np

from . import spatial_types as cas
from .degree_of_freedom import DegreeOfFreedom
from .prefixed_name import PrefixedName
from .spatial_types.derivatives import DerivativeMap
from .spatial_types.math import quaternion_from_rotation_matrix
from .types import NpMatrix4x4
from .world_entity import Connection

if TYPE_CHECKING:
    from .world import World


class Has1DOFState:
    """
    Mixin class that implements state access for connections with 1 degree of freedom.
    """

    dof: DegreeOfFreedom
    _world: World

    @property
    def position(self) -> float:
        return self._world.state[self.dof.name].position

    @position.setter
    def position(self, value: float) -> None:
        self._world.state[self.dof.name].position = value
        self._world.notify_state_change()

    @property
    def velocity(self) -> float:
        return self._world.state[self.dof.name].velocity

    @velocity.setter
    def velocity(self, value: float) -> None:
        self._world.state[self.dof.name].velocity = value
        self._world.notify_state_change()

    @property
    def acceleration(self) -> float:
        return self._world.state[self.dof.name].acceleration

    @acceleration.setter
    def acceleration(self, value: float) -> None:
        self._world.state[self.dof.name].acceleration = value
        self._world.notify_state_change()

    @property
    def jerk(self) -> float:
        return self._world.state[self.dof.name].jerk

    @jerk.setter
    def jerk(self, value: float) -> None:
        self._world.state[self.dof.name].jerk = value
        self._world.notify_state_change()


class HasUpdateState(ABC):
    """
    Mixin class for connections that need state updated which are not trivial integrations.
    Typically needed for connections that use active and passive degrees of freedom.
    Look at OmniDrive for an example usage.
    """

    @abstractmethod
    def update_state(self, dt: float) -> None:
        """
        Allows the connection to update the state of its dofs.
        An integration update for active dofs will have happened before this method is called.
        Write directly into self._world.state, but don't touch dofs that don't belong to this connection.
        :param dt: Time passed since last update.
        """
        pass


@dataclass
class FixedConnection(Connection):
    """
    Has 0 degrees of freedom.
    """


@dataclass
class ActiveConnection(Connection):
    """
    Has one or more degrees of freedom that can be actively controlled, e.g., robot joints.
    """

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return []


@dataclass
class PassiveConnection(Connection):
    """
    Has one or more degrees of freedom that cannot be actively controlled.
    Useful if a transformation is only tracked, e.g., the robot's localization.
    """

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return []


@dataclass
class PrismaticConnection(ActiveConnection, Has1DOFState):
    """
    Allows the movement along an axis.
    """

    axis: cas.Vector3 = field(kw_only=True)
    """
    Connection moves along this axis, should be a unit vector.
    The axis is defined relative to the local reference frame of the parent body.
    """

    multiplier: float = 1.0
    """
    Movement along the axis is multiplied by this value. Useful if Connections share DoFs.
    """

    offset: float = 0.0
    """
    Movement along the axis is offset by this value. Useful if Connections share DoFs.
    """

    dof: DegreeOfFreedom = field(default=None)
    """
    Degree of freedom to control movement along the axis.
    """

    def __post_init__(self):
        super().__post_init__()
        if self.multiplier is None:
            self.multiplier = 1
        else:
            self.multiplier = self.multiplier
        if self.offset is None:
            self.offset = 0
        else:
            self.offset = self.offset
        self.axis = self.axis
        self._post_init_world_part()

        motor_expression = self.dof.symbols.position * self.multiplier + self.offset
        translation_axis = cas.Vector3.from_iterable(self.axis) * motor_expression
        parent_T_child = cas.TransformationMatrix.from_xyz_rpy(x=translation_axis[0],
                                                               y=translation_axis[1],
                                                               z=translation_axis[2])
        self.origin_expression = self.origin_expression.dot(parent_T_child)
        self.origin_expression.reference_frame = self.parent
        self.origin_expression.child_frame = self.child

    def _post_init_with_world(self):
        self.dof = self.dof or self._world.create_degree_of_freedom(name=PrefixedName(str(self.name)))

    def _post_init_without_world(self):
        if self.dof is None:
            raise ValueError("PrismaticConnection cannot be created without a world "
                             "if the dof is not provided.")

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return [self.dof]

    def __hash__(self):
        return hash((self.parent, self.child))


@dataclass
class RevoluteConnection(ActiveConnection, Has1DOFState):
    """
    Allows rotation about an axis.
    """

    axis: cas.Vector3 = field(kw_only=True)
    """
    Connection rotates about this axis, should be a unit vector.
    The axis is defined relative to the local reference frame of the parent body.
    """

    multiplier: float = 1.0
    """
    Rotation about the axis is multiplied by this value. Useful if Connections share DoFs.
    """

    offset: float = 0.0
    """
    Rotation about the axis is offset by this value. Useful if Connections share DoFs.
    """

    dof: DegreeOfFreedom = field(default=None)
    """
    Degree of freedom to control rotation about the axis.
    """

    def __post_init__(self):
        super().__post_init__()
        if self.multiplier is None:
            self.multiplier = 1
        else:
            self.multiplier = self.multiplier
        if self.offset is None:
            self.offset = 0
        else:
            self.offset = self.offset
        self.axis = self.axis
        self._post_init_world_part()

        motor_expression = self.dof.symbols.position * self.multiplier + self.offset
        parent_R_child = cas.RotationMatrix.from_axis_angle(self.axis, motor_expression)
        self.origin_expression = self.origin_expression @ cas.TransformationMatrix(parent_R_child)
        self.origin_expression.reference_frame = self.parent
        self.origin_expression.child_frame = self.child

    def _post_init_with_world(self):
        self.dof = self.dof or self._world.create_degree_of_freedom(name=PrefixedName(str(self.name)))

    def _post_init_without_world(self):
        if self.dof is None:
            raise ValueError("RevoluteConnection cannot be created without a world "
                             "if the dof is not provided.")

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return [self.dof]

    def __hash__(self):
        return hash((self.parent, self.child))


@dataclass
class Connection6DoF(PassiveConnection):
    """
    Has full 6 degrees of freedom, that cannot be actively controlled.
    Useful for synchronizing with transformations from external providers.
    """

    x: DegreeOfFreedom = field(default=None)
    """
    Displacement of child body with respect to parent body along the x-axis.
    """
    y: DegreeOfFreedom = field(default=None)
    """
    Displacement of child body with respect to parent body along the y-axis.
    """
    z: DegreeOfFreedom = field(default=None)
    """
    Displacement of child body with respect to parent body along the z-axis.
    """

    qx: DegreeOfFreedom = field(default=None)
    qy: DegreeOfFreedom = field(default=None)
    qz: DegreeOfFreedom = field(default=None)
    qw: DegreeOfFreedom = field(default=None)
    """
    Rotation of child body with respect to parent body represented as a quaternion.
    """

    def __post_init__(self):
        super().__post_init__()
        self._post_init_world_part()
        parent_P_child = cas.Point3(x=self.x.symbols.position,
                                    y=self.y.symbols.position,
                                    z=self.z.symbols.position)
        parent_R_child = cas.Quaternion(x=self.qx.symbols.position,
                                        y=self.qy.symbols.position,
                                        z=self.qz.symbols.position,
                                        w=self.qw.symbols.position).to_rotation_matrix()
        self.origin_expression = cas.TransformationMatrix.from_point_rotation_matrix(point=parent_P_child,
                                                                                     rotation_matrix=parent_R_child,
                                                                                     reference_frame=self.parent,
                                                                                     child_frame=self.child)

    def _post_init_with_world(self):
        if all(dof is None for dof in self.passive_dofs):
            self.x = self._world.create_degree_of_freedom(name=PrefixedName('x', str(self.name)))
            self.y = self._world.create_degree_of_freedom(name=PrefixedName('y', str(self.name)))
            self.z = self._world.create_degree_of_freedom(name=PrefixedName('z', str(self.name)))
            self.qx = self._world.create_degree_of_freedom(name=PrefixedName('qx', str(self.name)))
            self.qy = self._world.create_degree_of_freedom(name=PrefixedName('qy', str(self.name)))
            self.qz = self._world.create_degree_of_freedom(name=PrefixedName('qz', str(self.name)))
            self.qw = self._world.create_degree_of_freedom(name=PrefixedName('qw', str(self.name)))
            self._world.state[self.qw.name].position = 1.
        elif any(dof is None for dof in self.passive_dofs):
            raise ValueError("Connection6DoF can only be created "
                             "if you provide all or none of the passive degrees of freedom")

    def _post_init_without_world(self):
        if any(dof is None for dof in self.passive_dofs):
            raise ValueError("Connection6DoF cannot be created without a world "
                             "if some passive degrees of freedom are not provided.")

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x, self.y, self.z, self.qx, self.qy, self.qz, self.qw]

    @property
    def origin(self) -> cas.TransformationMatrix:
        return super().origin

    @origin.setter
    def origin(self, transformation: Union[NpMatrix4x4, cas.TransformationMatrix]) -> None:
        if isinstance(transformation, cas.TransformationMatrix):
            transformation = transformation.to_np()
        orientation = quaternion_from_rotation_matrix(transformation)
        self._world.state[self.x.name].position = transformation[0, 3]
        self._world.state[self.y.name].position = transformation[1, 3]
        self._world.state[self.z.name].position = transformation[2, 3]
        self._world.state[self.qx.name].position = orientation[0]
        self._world.state[self.qy.name].position = orientation[1]
        self._world.state[self.qz.name].position = orientation[2]
        self._world.state[self.qw.name].position = orientation[3]
        self._world.notify_state_change()


@dataclass
class OmniDrive(ActiveConnection, PassiveConnection, HasUpdateState):
    x: DegreeOfFreedom = field(default=None)
    y: DegreeOfFreedom = field(default=None)
    z: DegreeOfFreedom = field(default=None)
    roll: DegreeOfFreedom = field(default=None)
    pitch: DegreeOfFreedom = field(default=None)
    yaw: DegreeOfFreedom = field(default=None)
    x_vel: DegreeOfFreedom = field(default=None)
    y_vel: DegreeOfFreedom = field(default=None)

    translation_velocity_limits: float = field(default=0.6)
    rotation_velocity_limits: float = field(default=0.5)

    def __post_init__(self):
        super().__post_init__()
        self._post_init_world_part()
        odom_T_bf = cas.TransformationMatrix.from_xyz_rpy(x=self.x.symbols.position,
                                                          y=self.y.symbols.position,
                                                          yaw=self.yaw.symbols.position)
        bf_T_bf_vel = cas.TransformationMatrix.from_xyz_rpy(x=self.x_vel.symbols.position,
                                                            y=self.y_vel.symbols.position)
        bf_vel_T_bf = cas.TransformationMatrix.from_xyz_rpy(x=0,
                                                            y=0,
                                                            z=self.z.symbols.position,
                                                            roll=self.roll.symbols.position,
                                                            pitch=self.pitch.symbols.position,
                                                            yaw=0)
        self.origin_expression = odom_T_bf.dot(bf_T_bf_vel).dot(bf_vel_T_bf)
        self.origin_expression.reference_frame = self.parent
        self.origin_expression.child_frame = self.child

    def _post_init_with_world(self):
        if all(dof is None for dof in self.dofs):
            stringified_name = str(self.name)
            lower_translation_limits = DerivativeMap()
            lower_translation_limits.velocity = -self.translation_velocity_limits
            upper_translation_limits = DerivativeMap()
            upper_translation_limits.velocity = self.translation_velocity_limits
            lower_rotation_limits = DerivativeMap()
            lower_rotation_limits.velocity = -self.rotation_velocity_limits
            upper_rotation_limits = DerivativeMap()
            upper_rotation_limits.velocity = self.rotation_velocity_limits

            self.x = self._world.create_degree_of_freedom(name=PrefixedName('x', stringified_name))
            self.y = self._world.create_degree_of_freedom(name=PrefixedName('y', stringified_name))
            self.z = self._world.create_degree_of_freedom(name=PrefixedName('z', stringified_name))

            self.roll = self._world.create_degree_of_freedom(name=PrefixedName('roll', stringified_name))
            self.pitch = self._world.create_degree_of_freedom(name=PrefixedName('pitch', stringified_name))
            self.yaw = self._world.create_degree_of_freedom(
                name=PrefixedName('yaw', stringified_name),
                lower_limits=lower_rotation_limits,
                upper_limits=upper_rotation_limits)

            self.x_vel = self._world.create_degree_of_freedom(
                name=PrefixedName('x_vel', stringified_name),
                lower_limits=lower_translation_limits,
                upper_limits=upper_translation_limits)
            self.y_vel = self._world.create_degree_of_freedom(
                name=PrefixedName('y_vel', stringified_name),
                lower_limits=lower_translation_limits,
                upper_limits=upper_translation_limits)
        elif any(dof is None for dof in self.passive_dofs):
            raise ValueError("OmniDrive can only be created "
                             "if you provide all or none of the passive degrees of freedom")

    def _post_init_without_world(self):
        if any(dof is None for dof in self.dofs):
            raise ValueError("OmniDrive cannot be created without a world "
                             "if some passive degrees of freedom are not provided.")

    @property
    def active_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x_vel, self.y_vel, self.yaw]

    @property
    def passive_dofs(self) -> List[DegreeOfFreedom]:
        return [self.x, self.y, self.z, self.roll, self.pitch]

    @property
    def dofs(self) -> List[DegreeOfFreedom]:
        return self.active_dofs + self.passive_dofs

    def update_state(self, dt: float) -> None:
        state = self._world.state
        state[self.x_vel.name].position = 0
        state[self.y_vel.name].position = 0

        x_vel = state[self.x_vel.name].velocity
        y_vel = state[self.y_vel.name].velocity
        delta = state[self.yaw.name].position
        state[self.x.name].velocity = (np.cos(delta) * x_vel - np.sin(delta) * y_vel)
        state[self.x.name].position += state[self.x.name].velocity * dt
        state[self.y.name].velocity = (np.sin(delta) * x_vel + np.cos(delta) * y_vel)
        state[self.y.name].position += state[self.y.name].velocity * dt

    def get_free_variable_names(self) -> List[PrefixedName]:
        return [self.x.name, self.y.name, self.yaw.name]



from __future__ import annotations

import logging
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from typing import Tuple, Iterable, Set

from typing_extensions import Optional, List, Self

from .prefixed_name import PrefixedName
from .spatial_types.spatial_types import Vector3
from .world import World
from .world_entity import Body, RootedView, Connection


@dataclass
class RobotView(RootedView, ABC):
    """
    Represents a collection of connected robot bodies, starting from a root body, and ending in a unspecified collection
    of tip bodies.
    """
    _robot: AbstractRobot = field(default=None)
    """
    The robot this view belongs to
    """

    def __post_init__(self):
        if self._world is not None:
            self._world.add_view(self)

    @abstractmethod
    def assign_to_robot(self, robot: AbstractRobot):
        """
        This method assigns the robot to the current view, and then iterates through its own fields to call the
        appropriate methods to att them to the robot.

        :param robot: The robot to which this view should be assigned.
        """
        ...


@dataclass
class KinematicChain(RobotView, ABC):
    """
    Abstract base class for kinematic chain in a robot, starting from a root body, and ending in a specific tip body.
    A kinematic chain can contain both a manipulator and sensors at the same time. There are no assumptions about the
    position of the manipulator or sensors in the kinematic chain
    """
    tip: Body = field(default_factory=Body)
    """
    The tip body of the kinematic chain, which is the last body in the chain.
    """

    manipulator: Optional[Manipulator] = None
    """
    The manipulator of the kinematic chain, if it exists. This is usually a gripper or similar device.
    """

    sensors: Set[Sensor] = field(default_factory=set)
    """
    A collection of sensors in the kinematic chain, such as cameras or other sensors.
    """

    @property
    def bodies(self) -> Iterable[Body]:
        """
        Returns itself as a kinematic chain.
        """
        return self._world.compute_chain_of_bodies(self.root, self.tip)

    @property
    def connections(self) -> Iterable[Connection]:
        """
        Returns the connections of the kinematic chain.
        This is a list of connections between the bodies in the kinematic chain
        """
        return self._world.compute_chain_of_connections(self.root, self.tip)

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the kinematic chain to the given robot. This method ensures that the kinematic chain is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(f"Kinematic chain {self.name} is already part of another robot: {self._robot.name}.")
        if self._robot is not None and self._robot == robot:
            return
        self._robot = robot
        if self.manipulator is not None:
            robot.add_manipulator(self.manipulator)
        for sensor in self.sensors:
            robot.add_sensor(sensor)

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Arm(KinematicChain):
    """
    Represents an arm of a robot, which is a kinematic chain with a specific tip body.
    An arm has a manipulators and potentially sensors.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Manipulator(RobotView, ABC):
    """
    Abstract base class of robot manipulators. Always has a tool frame.
    """
    tool_frame: Body = field(default_factory=Body)

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the manipulator to the given robot. This method ensures that the manipulator is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(f"Manipulator {self.name} is already part of another robot: {self._robot.name}.")
        if self._robot is not None and self._robot == robot:
            return
        self._robot = robot

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tool_frame))


@dataclass
class Finger(KinematicChain):
    """
    A finger is a kinematic chain, since it should have an unambiguous tip body, and may contain sensors.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class ParallelGripper(Manipulator):
    """
    Represents a gripper of a robot. Contains a collection of fingers and a thumb. The thumb is a specific finger
    that always needs to touch an object when grasping it, ensuring a stable grasp.
    """
    finger: Finger = field(default_factory=Finger)
    thumb: Finger = field(default_factory=Finger)

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the parallel gripper to the given robot and calls the appropriate methods for the its finger and thumb.
         This method ensures that the parallel gripper is only assigned to one robot at a time, and raises an error if
         it is already assigned to another
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(f"ParallelGripper {self.name} is already part of another robot: {self._robot.name}.")
        if self._robot is not None and self._robot == robot:
            return
        self._robot = robot
        robot.add_kinematic_chain(self.finger)
        robot.add_kinematic_chain(self.thumb)

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tool_frame))


@dataclass
class Sensor(RobotView, ABC):
    """
    Abstract base class for any kind of sensor in a robot.
    """

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the sensor to the given robot. This method ensures that the sensor is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(f"Sensor {self.name} is already part of another robot: {self._robot.name}.")
        if self._robot is not None and self._robot == robot:
            return
        self._robot = robot


@dataclass
class FieldOfView:
    """
    Represents the field of view of a camera sensor, defined by the vertical and horizontal angles of the camera's view.
    """
    vertical_angle: float
    horizontal_angle: float


@dataclass
class Camera(Sensor):
    """
    Represents a camera sensor in a robot.
    """
    forward_facing_axis: Vector3 = field(default_factory=Vector3)
    field_of_view: FieldOfView = field(default_factory=FieldOfView)
    minimal_height: float = 0.0
    maximal_height: float = 1.0

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root))


@dataclass
class Neck(KinematicChain):
    """
    Represents a special kinematic chain that connects the head of a robot with a collection of sensors, such as cameras
    and which does not have a manipulator.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Torso(KinematicChain):
    """
    A Torso is a kinematic chain connecting the base of the robot with a collection of other kinematic chains.
    """

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the torso to the given robot and calls the appropriate method for each of its attached kinematic chains.
         This method ensures that the torso is only assigned to one robot at a time, and raises an error if it is
         already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(f"Torso {self.name} is already part of another robot: {self._robot.name}.")
        if self._robot is not None and self._robot == robot:
            return
        self._robot = robot

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class AbstractRobot(RootedView, ABC):
    """
    Specification of an abstract robot. A robot consists of:
    - a root body, which is the base of the robot
    - an optional torso, which is a kinematic chain (usually without a manipulator) connecting the base with a collection
        of other kinematic chains
    - an optional collection of manipulator chains, each containing a manipulator, such as a gripper
    - an optional collection of sensor chains, each containing a sensor, such as a camera
    => If a kinematic chain contains both a manipulator and a sensor, it will be part of both collections
    """
    odom: Body = field(default_factory=Body)
    """
    The odometry body of the robot, which is usually the base footprint.
    """

    torso: Optional[Torso] = None
    """
    The torso of the robot, which is a kinematic chain connecting the base with a collection of other kinematic chains.
    """

    manipulators: Set[Manipulator] = field(default_factory=set)
    """
    A collection of manipulators in the robot, such as grippers.
    """

    sensors: Set[Sensor] = field(default_factory=set)
    """
    A collection of sensors in the robot, such as cameras.
    """

    manipulator_chains: Set[KinematicChain] = field(default_factory=set)
    """
    A collection of all kinematic chains containing a manipulator, such as a gripper.
    """

    sensor_chains: Set[KinematicChain] = field(default_factory=set)
    """
    A collection of all kinematic chains containing a sensor, such as a camera.
    """

    @classmethod
    @abstractmethod
    def from_world(cls, world: World) -> Self:
        """
        Creates a robot view from the given world.
        This method constructs the robot view by identifying and organizing the various semantic components of the robot,
        such as manipulators, sensors, and kinematic chains. It is expected to be implemented in subclasses.

        :param world: The world from which to create the robot view.

        :return: A robot view.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")

    def add_manipulator(self, manipulator: Manipulator):
        """
        Adds a manipulator to the robot's collection of manipulators.
        """
        self.manipulators.add(manipulator)
        self._views.add(manipulator)
        manipulator.assign_to_robot(self)

    def add_sensor(self, sensor: Sensor):
        """
        Adds a sensor to the robot's collection of sensors.
        """
        self.sensors.add(sensor)
        self._views.add(sensor)
        sensor.assign_to_robot(self)

    def add_torso(self, torso: Torso):
        """
        Adds a torso to the robot's collection of kinematic chains.
        """
        if self.torso is not None:
            raise ValueError(f"Robot {self.name} already has a torso: {self.torso.name}.")
        self.torso = torso
        self._views.add(torso)
        torso.assign_to_robot(self)

    def add_kinematic_chain(self, kinematic_chain: KinematicChain):
        """
        Adds a kinematic chain to the robot's collection of kinematic chains.
        This can be either a manipulator chain or a sensor chain.
        """
        if kinematic_chain.manipulator is None and not kinematic_chain.sensors:
            logging.warning(
                f"Kinematic chain {kinematic_chain.name} has no manipulator or sensors, so it was skipped. Did you mean to add it to the torso?")
            return
        if kinematic_chain.manipulator is not None:
            self.manipulator_chains.add(kinematic_chain)
        if kinematic_chain.sensors:
            self.sensor_chains.add(kinematic_chain)
        self._views.add(kinematic_chain)
        kinematic_chain.assign_to_robot(self)


@dataclass
class PR2(AbstractRobot):
    """
    Represents the Personal Robot 2 (PR2), which was originally created by Willow Garage.
    The PR2 robot consists of two arms, each with a parallel gripper, a head with a camera, and a prismatic torso
    """
    neck: Neck = field(default_factory=Neck)
    left_arm: KinematicChain = field(default_factory=KinematicChain)
    right_arm: KinematicChain = field(default_factory=KinematicChain)

    def _add_arm(self, arm: KinematicChain, arm_side: str):
        """
        Adds a kinematic chain to the PR2 robot's collection of kinematic chains.
        If the kinematic chain is an arm, it will be added to the left or right arm accordingly.

        :param arm: The kinematic chain to add to the PR2 robot.
        """
        if arm.manipulator is None:
            raise ValueError(f"Arm kinematic chain {arm.name} must have a manipulator.")

        if arm_side == 'left':
            self.left_arm = arm
        elif arm_side == 'right':
            self.right_arm = arm
        else:
            raise ValueError(f"Invalid arm side: {arm_side}. Must be 'left' or 'right'.")

        super().add_kinematic_chain(arm)

    def add_left_arm(self, kinematic_chain: KinematicChain):
        """
        Adds a left arm kinematic chain to the PR2 robot.

        :param kinematic_chain: The kinematic chain representing the left arm.
        """
        self._add_arm(kinematic_chain, 'left')

    def add_right_arm(self, kinematic_chain: KinematicChain):
        """
        Adds a right arm kinematic chain to the PR2 robot.

        :param kinematic_chain: The kinematic chain representing the right arm.
        """
        self._add_arm(kinematic_chain, 'right')

    def add_neck(self, neck: Neck):
        """
        Adds a neck kinematic chain to the PR2 robot.

        :param neck: The neck kinematic chain to add.
        """
        if not neck.sensors:
            raise ValueError(f"Neck kinematic chain {neck.name} must have at least one sensor.")
        self.neck = neck
        super().add_kinematic_chain(neck)

    @classmethod
    def from_world(cls, world: World) -> Self:
        """
        Creates a PR2 robot view from the given world.

        :param world: The world from which to create the robot view.

        :return: A PR2 robot view.
        """

        robot = cls(
            name=PrefixedName(name='pr2', prefix=world.name),
            odom=world.get_body_by_name("odom_combined"),
            root=world.get_body_by_name("base_footprint"),
            _world=world,
        )

        # Create left arm
        left_gripper_thumb = Finger(name=PrefixedName('left_gripper_thumb', prefix=robot.name.name),
                                    root=world.get_body_by_name("l_gripper_l_finger_link"),
                                    tip=world.get_body_by_name("l_gripper_l_finger_tip_link"),
                                    _world=world)

        left_gripper_finger = Finger(name=PrefixedName('left_gripper_finger', prefix=robot.name.name),
                                     root=world.get_body_by_name("l_gripper_r_finger_link"),
                                     tip=world.get_body_by_name("l_gripper_r_finger_tip_link"),
                                     _world=world)

        left_gripper = ParallelGripper(name=PrefixedName('left_gripper', prefix=robot.name.name),
                                       root=world.get_body_by_name("l_gripper_palm_link"),
                                       tool_frame=world.get_body_by_name("l_gripper_tool_frame"),
                                       thumb=left_gripper_thumb,
                                       finger=left_gripper_finger,
                                       _world=world)

        left_arm = Arm(name=PrefixedName('left_arm', prefix=robot.name.name),
                       root=world.get_body_by_name("l_shoulder_pan_link"),
                       manipulator=left_gripper,
                       _world=world)

        robot.add_left_arm(left_arm)

        # Create right arm
        right_gripper_thumb = Finger(name=PrefixedName('right_gripper_thumb', prefix=robot.name.name),
                                     root=world.get_body_by_name("r_gripper_l_finger_link"),
                                     tip=world.get_body_by_name("r_gripper_l_finger_tip_link"),
                                     _world=world)
        right_gripper_finger = Finger(name=PrefixedName('right_gripper_finger', prefix=robot.name.name),
                                      root=world.get_body_by_name("r_gripper_r_finger_link"),
                                      tip=world.get_body_by_name("r_gripper_r_finger_tip_link"),
                                      _world=world)
        right_gripper = ParallelGripper(name=PrefixedName('right_gripper', prefix=robot.name.name),
                                        root=world.get_body_by_name("r_gripper_palm_link"),
                                        tool_frame=world.get_body_by_name("r_gripper_tool_frame"),
                                        thumb=right_gripper_thumb,
                                        finger=right_gripper_finger,
                                        _world=world)
        right_arm = Arm(name=PrefixedName('right_arm', prefix=robot.name.name),
                        root=world.get_body_by_name("r_shoulder_pan_link"),
                        manipulator=right_gripper,
                        _world=world)

        robot.add_right_arm(right_arm)

        # Create camera and neck
        camera = Camera(name=PrefixedName('wide_stereo_optical_frame', prefix=robot.name.name),
                        forward_facing_axis=Vector3(0, 0, 1),
                        field_of_view=FieldOfView(horizontal_angle=0.99483, vertical_angle=0.75049),
                        minimal_height=1.27,
                        maximal_height=1.60,
                        _world=world)

        neck = Neck(name=PrefixedName('neck', prefix=robot.name.name),
                    sensors={camera},
                    root=world.get_body_by_name("head_pan_link"),
                    tip=world.get_body_by_name("head_tilt_link"),
                    _world=world)
        robot.add_neck(neck)

        # Create torso
        torso = Torso(name=PrefixedName('torso', prefix=robot.name.name),
                      root=world.get_body_by_name("torso_lift_link"),
                      tip=world.get_body_by_name("torso_lift_link"),
                      _world=world)
        robot.add_torso(torso)

        world.add_view(robot)

        return robot



from __future__ import annotations

import inspect
from abc import abstractmethod
from collections import deque
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field, fields
from functools import reduce
from typing import List, Optional, TYPE_CHECKING, Set, get_args, get_type_hints, Deque
import numpy as np
from numpy import ndarray
from typing_extensions import Self

from .geometry import Shape, BoundingBox, BoundingBoxCollection
from dataclasses import dataclass, field
from dataclasses import fields
from functools import lru_cache
from functools import reduce
from typing import List, Optional, TYPE_CHECKING, Tuple
from typing import Set

import numpy as np
from numpy import ndarray
from scipy.stats import geom
from trimesh.proximity import closest_point, nearby_faces
from trimesh.sample import sample_surface

from .geometry import BoundingBox, BoundingBoxCollection
from .geometry import Shape
from .prefixed_name import PrefixedName
from .spatial_types import spatial_types as cas
from .spatial_types.spatial_types import Point3
from .spatial_types.spatial_types import TransformationMatrix, Expression
from .types import NpMatrix4x4
from .utils import IDGenerator

if TYPE_CHECKING:
    from .world import World

id_generator = IDGenerator()


@dataclass(unsafe_hash=True)
class WorldEntity:
    """
    A class representing an entity in the world.
    """

    _world: Optional[World] = field(default=None, repr=False, kw_only=True, hash=False)
    """
    The backreference to the world this entity belongs to.
    """

    _views: Set[View] = field(default_factory=set, init=False, repr=False, hash=False)
    """
    The views this entity is part of.
    """

    name: PrefixedName = field(default=None, kw_only=True)
    """
    The identifier for this world entity.
    """

    def __post_init__(self):
        if self.name is None:
            self.name = PrefixedName(f"{self.__class__.__name__}_{hash(self)}")


@dataclass
class Body(WorldEntity):
    """
    Represents a body in the world.
    A body is a semantic atom, meaning that it cannot be decomposed into meaningful smaller parts.
    """

    visual: List[Shape] = field(default_factory=list, repr=False)
    """
    List of shapes that represent the visual appearance of the link.
    The poses of the shapes are relative to the link.
    """

    collision: List[Shape] = field(default_factory=list, repr=False)
    """
    List of shapes that represent the collision geometry of the link.
    The poses of the shapes are relative to the link.
    """

    index: Optional[int] = field(default=None, init=False)
    """
    The index of the entity in `_world.kinematic_structure`.
    """

    def __post_init__(self):
        if not self.name:
            self.name = PrefixedName(f"body_{id_generator(self)}")

        if self._world is not None:
            self.index = self._world.kinematic_structure.add_node(self)

        for c in self.collision:
            c.origin.reference_frame = self
        for v in self.visual:
            v.origin.reference_frame = self

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name and self._world is other._world

    def has_collision(self) -> bool:
        return len(self.collision) > 0

    def compute_closest_points_multi(self, others: list[Body], sample_size=25) -> Tuple[
        np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the closest points to each given body respectively.

        :param others: The list of bodies to compute the closest points to.
        :param sample_size: The number of samples to take from the surface of the other bodies.
        :return: A tuple containing: The points on the self body, the points on the other bodies, and the distances. All points are in the of this body.
        """

        @lru_cache(maxsize=None)
        def evaluated_geometric_distribution(n: int) -> np.ndarray:
            """
            Evaluates the geometric distribution for a given number of samples.
            :param n: The number of samples to evaluate.
            :return: An array of probabilities for each sample.
            """
            return geom.pmf(np.arange(1, n + 1), 0.5)

        query_points = []
        for other in others:
            # Calculate the closest vertex on this body to the other body
            closest_vert_id = \
                self.collision[0].mesh.kdtree.query(
                    (self._world.compute_forward_kinematics_np(self, other) @ other.collision[0].origin.to_np())[:3, 3],
                    k=1)[1]
            closest_vert = self.collision[0].mesh.vertices[closest_vert_id]

            # Compute the closest faces on the other body to the closes vertex
            faces = nearby_faces(other.collision[0].mesh,
                                 [(self._world.compute_forward_kinematics_np(other, self) @ self.collision[
                                     0].origin.to_np())[:3, 3] + closest_vert])[0]
            face_weights = np.zeros(len(other.collision[0].mesh.faces))

            # Assign weights to the faces based on a geometric distribution
            face_weights[faces] = evaluated_geometric_distribution(len(faces))

            # Sample points on the surface of the other body
            q = sample_surface(other.collision[0].mesh, sample_size, face_weight=face_weights, seed=420)[0]
            # Make 4x4 transformation matrix from points
            points = np.tile(np.eye(4, dtype=np.float32), (len(q), 1, 1))
            points[:, :3, 3] = q

            # Transform from the mesh to the other mesh
            transform = np.linalg.inv(self.collision[0].origin.to_np()) @ self._world.compute_forward_kinematics_np(
                self, other) @ other.collision[0].origin.to_np()
            points = points @ transform

            points = points[:, :3, 3]  # Extract the points from the transformation matrix

            query_points.extend(points)

        # Actually compute the closest points
        points, dists = closest_point(self.collision[0].mesh, query_points)[:2]
        # Find the closest points for each body out of all the sampled points
        points = np.array(points).reshape(len(others), sample_size, 3)
        dists = np.array(dists).reshape(len(others), sample_size)
        dist_min = np.min(dists, axis=1)
        points_min_self = points[np.arange(len(others)), np.argmin(dists, axis=1), :]
        points_min_other = np.array(query_points).reshape(len(others), sample_size, 3)[np.arange(len(others)),
                           np.argmin(dists, axis=1), :]
        return points_min_self, points_min_other, dist_min

    @property
    def child_bodies(self) -> List[Body]:
        """
        Returns the child bodies of this body.
        """
        return self._world.compute_child_bodies(self)

    @property
    def parent_body(self) -> Body:
        """
        Returns the parent body of this body.
        """
        return self._world.compute_parent_body(self)

    @property
    def bounding_box_collection(self) -> BoundingBoxCollection:
        """
        Return the bounding box collection of the link with the given name.
        This method computes the bounding box of the link in world coordinates by transforming the local axis-aligned
        bounding boxes of the link's geometry to world coordinates.

        Note: These bounding boxes may not be disjoint, however the random events library always makes them disjoint. If
        this is the case, and we feed he non-disjoint bounding boxes into the gcs, it may trigger unexpected behavior.

        :return: A BoundingBoxCollection containing the bounding boxes of the link's geometry in world
        """
        world = self._world
        body_transform: ndarray = world.compute_forward_kinematics_np(world.root, self)
        world_bboxes = []

        for shape in self.collision:
            shape_transform: ndarray = shape.origin.to_np()

            world_transform: ndarray = body_transform @ shape_transform
            body_pos = world_transform[:3, 3]
            body_rotation_matrix = world_transform[:3, :3]

            local_bb: BoundingBox = shape.as_bounding_box()

            # Get all 8 corners of the BB in link-local space
            corners = np.array([corner.to_np()[:3] for corner in local_bb.get_points()])  # shape (8, 3)

            # Transform each corner to world space: R * corner + T
            transformed_corners = (corners @ body_rotation_matrix.T) + body_pos

            # Compute world-space bounding box from transformed corners
            min_corner = np.min(transformed_corners, axis=0)
            max_corner = np.max(transformed_corners, axis=0)

            world_bb = BoundingBox.from_min_max(Point3(*min_corner), Point3(*max_corner))
            world_bboxes.append(world_bb)

        return BoundingBoxCollection(world_bboxes)

    @property
    def global_pose(self) -> NpMatrix4x4:
        """
        Computes the pose of the body in the world frame.
        :return: 4x4 transformation matrix.
        """
        return self._world.compute_forward_kinematics_np(self._world.root, self)

    @property
    def parent_connection(self) -> Connection:
        """
        Returns the parent connection of this body.
        """
        return self._world.compute_parent_connection(self)

    @classmethod
    def from_body(cls, body: Body):
        """
        Creates a new link from an existing link.
        """
        new_link = cls(name=body.name, visual=body.visual, collision=body.collision)
        new_link._world = body._world
        new_link.index = body.index
        return new_link


@dataclass
class View(WorldEntity):
    """
    Represents a view on a set of bodies in the world.

    This class can hold references to certain bodies that gain meaning in this context.
    """

    def _bodies(self, visited: Set[int]) -> Set[Body]:
        """
        Recursively collects all bodies that are part of this view.
        """
        stack: Deque[object] = deque([self])
        bodies: Set[Body] = set()

        while stack:
            obj = stack.pop()
            oid = id(obj)
            if oid in visited:
                continue
            visited.add(oid)

            match obj:
                case Body():
                    bodies.add(obj)

                case View():
                    stack.extend(_attr_values(obj))

                case Mapping():
                    stack.extend(v for v in obj.values() if _is_body_view_or_iterable(v))

                case Iterable() if not isinstance(obj, (str, bytes, bytearray)):
                    stack.extend(v for v in obj if _is_body_view_or_iterable(v))

        return bodies

    @property
    def bodies(self) -> Iterable[Body]:
        """
        Returns a Iterable of all relevant bodies in this view. The default behaviour is to aggregate all bodies that are accessible
        through the properties and fields of this view, recursively.
        If this behaviour is not desired for a specific view, it can be overridden by implementing the `bodies` property.
        """
        return self._bodies(set())

    def as_bounding_box_collection(self) -> BoundingBoxCollection:
        """
        Returns a bounding box collection that contains the bounding boxes of all bodies in this view.
        """
        bbs = reduce(
            lambda accumulator, bb_collection: accumulator.merge(bb_collection),
            (body.bounding_box_collection for body in self.bodies if body.has_collision())
        )
        return bbs


@dataclass(unsafe_hash=True)
class RootedView(View):
    """
    Represents a view that is rooted in a specific body.
    """
    root: Body = field(default_factory=Body)


@dataclass(unsafe_hash=True)
class EnvironmentView(RootedView):
    """
    Represents a view of the environment.
    """

    @property
    def bodies(self) -> Set[Body]:
        """
        Returns a set of all bodies in the environment view.
        """
        return set(self._world.compute_child_bodies_recursive(self.root)) | {self.root}


@dataclass
class Connection(WorldEntity):
    """
    Represents a connection between two bodies in the world.
    """

    parent: Body
    """
    The parent body of the connection.
    """

    child: Body
    """
    The child body of the connection.
    """

    origin_expression: TransformationMatrix = field(default=None)
    """
    A symbolic expression describing the origin of the connection.
    """

    def __post_init__(self):
        if self.origin_expression is None:
            self.origin_expression = TransformationMatrix()
        self.origin_expression.reference_frame = self.parent
        self.origin_expression.child_frame = self.child
        if self.name is None:
            self.name = PrefixedName(f'{self.parent.name.name}_T_{self.child.name.name}', prefix=self.child.name.prefix)

    def _post_init_world_part(self):
        """
        Executes post-initialization logic based on the presence of a world attribute.
        """
        if self._world is None:
            self._post_init_without_world()
        else:
            self._post_init_with_world()

    def _post_init_with_world(self):
        """
        Initialize or perform additional setup operations required after the main
        initialization step. Use for world-related configurations or specific setup
        details required post object creation.
        """
        pass

    def _post_init_without_world(self):
        """
        Handle internal initialization processes when _world is None. Perform
        operations post-initialization for internal use only.
        """
        pass

    def __hash__(self):
        return hash((self.parent, self.child))

    def __eq__(self, other):
        return self.name == other.name

    @property
    def origin(self) -> cas.TransformationMatrix:
        """
        :return: The relative transform between the parent and child frame.
        """
        return self._world.compute_forward_kinematics(self.parent, self.child)

    # @lru_cache(maxsize=None)
    def origin_as_position_quaternion(self) -> Expression:
        position = self.origin_expression.to_position()[:3]
        orientation = self.origin_expression.to_quaternion()
        return cas.vstack([position, orientation]).T


def _is_body_view_or_iterable(obj: object) -> bool:
    """
    Determines if an object is a Body, a View, or an Iterable (excluding strings and bytes).
    """
    return (
            isinstance(obj, (Body, View)) or
            (isinstance(obj, Iterable) and not isinstance(obj, (str, bytes, bytearray)))
    )


def _attr_values(view: View) -> Iterable[object]:
    """
    Yields all dataclass fields and set properties of this view.
    Skips private fields (those starting with '_'), as well as the 'bodies' property.

    :param view: The view to extract attributes from.
    """
    for f in fields(view):
        if f.name.startswith('_'):
            continue
        v = getattr(view, f.name, None)
        if _is_body_view_or_iterable(v):
            yield v

    for name, prop in inspect.getmembers(type(view), lambda o: isinstance(o, property)):
        if name == "bodies" or name.startswith('_'):
            continue
        try:
            v = getattr(view, name)
        except Exception:
            continue
        if _is_body_view_or_iterable(v):
            yield v



from dataclasses import dataclass, field
from typing import List, Optional

from ormatic.dao import AlternativeMapping, T

from ..degree_of_freedom import DegreeOfFreedom
from ..prefixed_name import PrefixedName
from ..spatial_types import RotationMatrix, Vector3, Point3, TransformationMatrix
from ..spatial_types.derivatives import DerivativeMap
from ..spatial_types.spatial_types import Quaternion
from ..spatial_types.symbol_manager import symbol_manager
from ..world import World, Body
from ..world_entity import Connection, View


@dataclass
class WorldMapping(AlternativeMapping[World]):
    bodies: List[Body]
    connections: List[Connection]
    views: List[View]
    degrees_of_freedom: List[DegreeOfFreedom]

    @classmethod
    def create_instance(cls, obj: World):
        # return cls(obj.bodies[:2], [],[],[], )
        return cls(obj.bodies, obj.connections, obj.views, list(obj.degrees_of_freedom))

    def create_from_dao(self) -> World:
        result = World()

        with result.modify_world():
            for body in self.bodies:
                result.add_body(body)
            for connection in self.connections:
                result.add_connection(connection)
            for view in self.views:
                result.add_view(view)
            for dof in self.degrees_of_freedom:
                result.create_degree_of_freedom(name=dof.name, lower_limits=dof.lower_limits,
                                                upper_limits=dof.upper_limits)

        return result


@dataclass
class Vector3Mapping(AlternativeMapping[Vector3]):
    x: float
    y: float
    z: float

    reference_frame: Optional[Body] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: Vector3):
        x, y, z, _ = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Vector3:
        return Vector3(x=self.x, y=self.y, z=self.z, reference_frame=None)



@dataclass
class Point3Mapping(AlternativeMapping[Point3]):
    x: float
    y: float
    z: float

    reference_frame: Optional[Body] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: Point3):
        x, y, z, _ = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Point3:
        return Point3(x=self.x, y=self.y, z=self.z, reference_frame=None)


@dataclass
class QuaternionMapping(AlternativeMapping[Quaternion]):
    x: float
    y: float
    z: float
    w: float

    reference_frame: Optional[Body] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: Quaternion):
        x, y, z, w = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z, w=w)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Quaternion:
        return Quaternion(x=self.x, y=self.y, z=self.z, w=self.w, reference_frame=None)


@dataclass
class RotationMatrixMapping(AlternativeMapping[RotationMatrix]):
    rotation: Quaternion
    reference_frame: Optional[Body] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: RotationMatrix):
        result = cls(rotation=obj.to_quaternion())
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> RotationMatrix:
        result = RotationMatrix.from_quaternion(self.rotation)
        result.reference_frame = None
        return result


@dataclass
class TransformationMatrixMapping(AlternativeMapping[TransformationMatrix]):
    position: Point3
    rotation: Quaternion
    reference_frame: Optional[Body] = field(init=False, default=None)
    child_frame: Optional[Body] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: TransformationMatrix):
        position = obj.to_position()
        rotation = obj.to_quaternion()
        result = cls(position=position,
                     rotation=rotation)
        result.reference_frame = obj.reference_frame
        result.child_frame = obj.child_frame

        return result

    def create_from_dao(self) -> TransformationMatrix:
        return TransformationMatrix.from_point_rotation_matrix(
            point=self.position,
            rotation_matrix=RotationMatrix.from_quaternion(self.rotation), reference_frame=None,
            child_frame=self.child_frame, )


@dataclass
class DegreeOfFreedomMapping(AlternativeMapping[DegreeOfFreedom]):
    name: PrefixedName
    lower_limits: List[float]
    upper_limits: List[float]

    @classmethod
    def create_instance(cls, obj: DegreeOfFreedom):
        return cls(name=obj.name, lower_limits=obj.lower_limits.data, upper_limits=obj.upper_limits.data)

    def create_from_dao(self) -> DegreeOfFreedom:
        lower_limits = DerivativeMap(data=self.lower_limits)
        upper_limits = DerivativeMap(data=self.upper_limits)
        return DegreeOfFreedom(name=self.name, lower_limits=lower_limits, upper_limits=upper_limits)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join("..", "src", "semantic_world")))

# -- Project information -----------------------------------------------------

project = 'semantic_world'
copyright = '2025, AICOR Institute for Artificial Intelligence'
author = 'AICOR Institute for Artificial Intelligence'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = ''